package org.jastadd.ros2rag.compiler;

import beaver.Parser;
import org.jastadd.ros2rag.ast.*;
import org.jastadd.ros2rag.compiler.options.CommandLine;
import org.jastadd.ros2rag.compiler.options.CommandLine.CommandLineException;
import org.jastadd.ros2rag.compiler.options.FlagOption;
import org.jastadd.ros2rag.compiler.options.Option;
import org.jastadd.ros2rag.compiler.options.StringOption;
import org.jastadd.ros2rag.parser.Ros2RagParser;
import org.jastadd.ros2rag.scanner.Ros2RagScanner;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.ArrayList;
import java.util.List;

public class Compiler {

  private StringOption optionOutputDir;
  private StringOption optionInputGrammar;
  private StringOption optionRootNode;
  private StringOption optionInputRos2Rag;
  private FlagOption optionHelp;
  private FlagOption optionVerbose;
  private FlagOption optionLogReads;
  private FlagOption optionLogWrites;

  private ArrayList<Option<?>> options;
  private CommandLine commandLine;

  public Compiler() {
    options = new ArrayList<>();
    addOptions();
  }

  public void run(String[] args) throws CommandLineException, CompilerException {
    System.setProperty("java.util.logging.manager", "org.apache.logging.log4j.jul.LogManager");
    System.setProperty("mustache.debug", "true");
    options = new ArrayList<>();
    addOptions();
    commandLine = new CommandLine(options);
    commandLine.parse(args);

    if (optionHelp.isSet()) {
      printUsage();
      return;
    }

    if (optionVerbose.isSet()) {
      try {
        run();
      } catch (CompilerException e) {
        e.printStackTrace();
        throw e;
      }
    } else {
      run();
    }
  }

  private void run() throws CompilerException {
    String outputDir;
    if (optionOutputDir.isSet()) {
      outputDir = optionOutputDir.getValue();
    } else {
      outputDir = "gen";
      System.out.println("No output output dir is set. Assuming '" + outputDir + "'.");
    }
    try {
      Files.createDirectories(Paths.get(outputDir));
    } catch (IOException e) {
      throw new CompilerException("Error creating output dir " + outputDir, e);
    }

    printMessage("Running Ros2Rag Preprocessor");

    if (anyRequiredOptionIsUnset()) {
      throw new CompilerException("Aborting due to missing values for required options.");
    }

    List<String> otherArgs = commandLine.getArguments();
    if (!otherArgs.isEmpty()) {
      printMessage("Superfluous arguments will be ignored: " + otherArgs);
    }
    Ros2Rag ros2Rag = parseProgram(optionInputGrammar.getValue(), optionInputRos2Rag.getValue());

    if (!ros2Rag.errors().isEmpty()) {
      System.err.println("Errors:");
      for (ErrorMessage e : ros2Rag.errors()) {
        System.err.println(e);
      }
      System.exit(1);
    }

    printMessage("Writing output files");
    // copy MqttHandler into outputDir
    final String mqttHandlerFileName = "MqttHandler.jadd";
    try {
      InputStream inputStream = Compiler.class.getClassLoader().getResourceAsStream(mqttHandlerFileName);
      if (inputStream == null) {
        throw new CompilerException("Could not open " + mqttHandlerFileName);
      }
      Files.copy(inputStream,
          Paths.get(outputDir, mqttHandlerFileName),
          StandardCopyOption.REPLACE_EXISTING);
    } catch (IOException e) {
      throw new CompilerException("Could not copy " + mqttHandlerFileName, e);
    }
    writeToFile(outputDir + "/Grammar.relast", ros2Rag.getProgram().generateAbstractGrammar());
    writeToFile(outputDir + "/ROS2RAG.jadd", ros2Rag.generateAspect(optionRootNode.getValue()));
  }

  private boolean anyRequiredOptionIsUnset() {
    boolean foundError = false;
    for (Option<?> option : options) {
      if (option.hasArgument() == Option.HasArgument.YES && !option.isSet()) {
        System.err.println("Option '" + option.getName() +
            "' (" + option.getDescription() + ") is required but unset!");
        foundError = true;
      }
    }
    return foundError;
  }


  public static void main(String[] args) {
    try {
      new Compiler().run(args);
    } catch (CommandLineException | CompilerException e) {
      System.err.println(e.getMessage());
      System.exit(-1);
    }
  }

  private void printMessage(String message) {
    System.out.println(message);
  }

  private void writeToFile(String filename, String str) throws CompilerException {
    try {
      PrintWriter writer = new PrintWriter(filename);
      writer.print(str);
      writer.close();
    } catch (Exception e) {
      throw new CompilerException("Could not write to file " + filename, e);
    }
  }

  private void addOptions() {
    optionOutputDir = addOption(new StringOption("outputDir", "target directory for the generated files."));
    optionInputGrammar = addOption(new StringOption("inputGrammar", "base grammar."));
    optionRootNode = addOption(new StringOption("rootNode", "root node in the base grammar."));
    optionInputRos2Rag = addOption(new StringOption("inputRos2Rag", "ros2rag definition file."));
    optionHelp = addOption(new FlagOption("help", "Print usage and exit."));
    optionVerbose = addOption(new FlagOption("verbose", "Print more messages while compiling."));
    optionLogReads = addOption(new FlagOption("logReads", "Enable logging for every read."));
    optionLogWrites = addOption(new FlagOption("logWrites", "Enable logging for every write."));
  }

  private <OptionType extends Option<?>> OptionType addOption(OptionType option) {
    options.add(option);
    return option;
  }

  private Ros2Rag parseProgram(String inputGrammarFileName, String inputRos2RagFileName) throws CompilerException {
    Program program = new Program();
    Ros2Rag ros2Rag;
    GrammarFile inputGrammar;

    try (BufferedReader reader = Files.newBufferedReader(Paths.get(inputGrammarFileName))) {
      Ros2RagScanner scanner = new Ros2RagScanner(reader);
      Ros2RagParser parser = new Ros2RagParser();
      inputGrammar = (GrammarFile) parser.parse(scanner);
      if (optionVerbose.isSet()) {
        inputGrammar.dumpTree(System.out);
      }
      program.addGrammarFile(inputGrammar);
      inputGrammar.treeResolveAll();
      inputGrammar.setFileName(inputGrammarFileName);
    } catch (IOException | Parser.Exception e) {
      throw new CompilerException("Could not parse grammar file " + inputGrammarFileName, e);
    }

    try (BufferedReader reader = Files.newBufferedReader(Paths.get(inputRos2RagFileName))) {
      Ros2RagScanner scanner = new Ros2RagScanner(reader);
      Ros2RagParser parser = new Ros2RagParser();
      ros2Rag = (Ros2Rag) parser.parse(scanner, Ros2RagParser.AltGoals.ros2rag);
      ros2Rag.setProgram(program);
      ros2Rag.setFileName(inputRos2RagFileName);
    } catch (IOException | Parser.Exception e) {
      throw new CompilerException("Could not parse ros2rag file " + inputRos2RagFileName, e);
    }
    ros2Rag.treeResolveAll();
    ros2Rag.additionalRelations().forEach(inputGrammar::addDeclaration);
    ASTNode.loggingEnabledForReads = optionLogReads.isSet();
    ASTNode.loggingEnabledForWrites = optionLogWrites.isSet();
    return ros2Rag;
  }

  protected void printUsage() {
    System.out.println("Usage: java -jar ros2rag.jar [--option1] [--option2=value] ...  <filename1> <filename2> ... ");
    System.out.println("Options:");
    System.out.print(commandLine.printOptionHelp());
  }

  public static class CompilerException extends Exception {
    CompilerException(String message) {
      super(message);
    }
    CompilerException(String message, Throwable cause) {
      super(message, cause);
    }
  }
}
